/* Copyright 2005 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: GameState.cpp,v 1.7 2005/04/19 19:22:46 tonic Exp $
 * $Revision: 1.7 $
 */

#include <stdio.h>
#include <math.h>

#include "main.h"
#include "Levels.h"
#include "GameState.h"


enum TILE_MAP_INFO_ENUM
{
    MAP_FILE_LAYERS = 2,
    TILE_WIDTH = 32,
    TILE_HEIGHT = 32,
    LIGHT_IMAGES = 9,
    MAX_LIGHT_IMAGES = (1 << 4),
    RELOAD_BULLETS = 8,
    RELOAD_AREA_HEIGHT = 32,
    LEVEL_START_TEXT_TIME = 3000,
    TILE_LIGHT_RENDERING_EXTRA_BORDER = 8,
    GUN_RELOAD_TIME = 450
};


class GameState::Light
{
public:
    Light() : mImage(NULL) {}
    ~Light() { delete mImage; }

    BOOL init(INT score, char *filename, FLOAT xOffset, FLOAT yOffset,
              FLOAT xScale, FLOAT yScale)
    {
        UIGraphics *g = GameApp::getInstance()->getGraphics();
        mImage = (UIImageOpenGL *)g->loadImage(filename);
        if (mImage != NULL)
        {
            mScore = score;
            mXOffset = xOffset * mImage->getWidth() * xScale;
            mYOffset = yOffset * mImage->getHeight() * yScale;
            mWidth = xScale * mImage->getWidth();
            mHeight = yScale * mImage->getHeight();
            glBindTexture(GL_TEXTURE_2D, mImage->getGLId());
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        }
        return mImage != NULL;
    }

    UIImageOpenGL *mImage;
    FLOAT mXOffset, mYOffset;
    FLOAT mWidth, mHeight;
    INT32 mScore;
};


GameState::GameState(UIGraphics *graphics) :
    mLights(NULL),
    mMapWidth(0), mMapHeight(0),
    mTileMap(NULL),
    mTileImage(NULL),
    mBulletImage(NULL),
    mGraphics(graphics),
    mButtonUpImage(NULL), mButtonDownImage(NULL),
    mSmallButtonUpImage(NULL), mSmallButtonDownImage(NULL)
{
}


GameState::~GameState()
{
}


BOOL GameState::init()
{
    BOOL success = TRUE;

    mLevels.init(this);
    success = mParticleSystem.init();

    GameApp *ga = GameApp::getInstance();
    INT scrW = ga->getScreenWidth(), scrH = ga->getScreenHeight();
    UIGraphics *graphics = ga->getGraphics();
    mUI.setGraphics(graphics);

    mLights = new Light[LIGHT_IMAGES];
    if (mLights == NULL)
        success = FALSE;

    if (success)
    {
        success &= mLights[0].init(10, DATA_PREFIX "light-big.png", -0.5f, -0.5f, 2, 2);
        success &= mLights[1].init(15, DATA_PREFIX "light-lamp.png", -0.5f, -0.1656f, 2, 2);
        success &= mLights[2].init(20, DATA_PREFIX "light-grating.png", -0.5f, -0.5, 3, 3);
        success &= mLights[3].init(50, DATA_PREFIX "light-candle.png", -1/3.f, -2/3.f, 1.5f, 1.5f);
        success &= mLights[4].init(30, DATA_PREFIX "light-exit.png", -0.5f, -0.5f, 8, 8);
        success &= mLights[5].init(40, DATA_PREFIX "light-lantern.png", -0.5f, -0.4f, 8, 8);
        success &= mLights[6].init(50, DATA_PREFIX "light-t.png", -0.5f, -0.5f, 4, 4);
        success &= mLights[7].init(50, DATA_PREFIX "light-A.png", -0.5f, -0.5f, 4, 4);
        success &= mLights[8].init(50, DATA_PREFIX "light-door.png", -0.5f, -1/3.f, 5, 5);
        mBulletImage = mGraphics->loadImage(DATA_PREFIX "bullet.png");
        if (mBulletImage == NULL)
            success = FALSE;
    }

    if (success)
    {
        mButtonUpImage = graphics->loadImage(DATA_PREFIX "button_up.png");
        mButtonDownImage = graphics->loadImage(DATA_PREFIX "button_down.png");
        mSmallButtonUpImage = graphics->loadImage(DATA_PREFIX "button_up_small.png");
        mSmallButtonDownImage = graphics->loadImage(DATA_PREFIX "button_down_small.png");
        if (!mButtonDownImage || !mButtonDownImage ||
            !mSmallButtonDownImage || !mSmallButtonDownImage)
            success = FALSE;
    }

    if (success)
    {
        UIFont *medium = ga->getMediumFont();
        UIFont *small = ga->getSmallFont();
        const UINT buttonWidth = mButtonUpImage->getWidth();
        const UINT buttonHeight = mButtonUpImage->getHeight();
        mContinueButton = new UIButton(scrW / 2 - buttonWidth / 2, scrH * 3 / 4,
                                       mButtonUpImage, mButtonDownImage, NULL,
                                       medium, "Continue");
        mContinueButton->addEventListener(this);
        mUI.addUIComponent(mContinueButton);
        if (!mContinueButton)
            success = FALSE;
    }

    return success;
}


void GameState::deinit()
{
    destroyMap();

    mUI.deleteAllUIComponents();

    delete mButtonUpImage;
    delete mButtonDownImage;
    delete mSmallButtonUpImage;
    delete mSmallButtonDownImage;

    delete mBulletImage;
    delete[] mLights;
    mLights = NULL;
}


BOOL GameState::loadMap(char *mapFilename, char *tileImageFilename)
{
    BOOL success = TRUE;

    destroyMap();

    FILE *fp = fopen(mapFilename, "rb");
    if (fp == NULL)
        return FALSE;

    fread(&mMapWidth, sizeof(UINT32), 1, fp);
    fread(&mMapHeight, sizeof(UINT32), 1, fp);
    mTileMap = new Tile[mMapWidth * mMapHeight];
    UINT16 *tempMap = new UINT16[mMapWidth * mMapHeight * MAP_FILE_LAYERS];
    if (mTileMap == NULL || tempMap == NULL)
    {
        errorMessage("Error", "Unable to allocate tile map (%d, %d)",
                     mMapWidth, mMapHeight);
        success = FALSE;
    }

    if (success)
    {
        INT layersRead;
        layersRead = fread(tempMap, mMapWidth * mMapHeight * sizeof(UINT16),
                           MAP_FILE_LAYERS, fp);
        if (layersRead != MAP_FILE_LAYERS)
        {
            errorMessage("Error", "Could read only %d layers from map file",
                         layersRead);
            success = FALSE;
        }
    }
    fclose(fp);

    if (success)
    {
        mLevelLightCount = 0;
        INT32 x, y, l;
        for (y = 0; y < mMapHeight; ++y)
        {
            for (x = 0; x < mMapWidth; ++x)
            {
                UINT16 layer[MAP_FILE_LAYERS];
                for (l = 0; l < MAP_FILE_LAYERS; ++l)
                    layer[l] = tempMap[l * mMapWidth * mMapHeight +
                                       y * mMapWidth + x];
                Tile &tile = mTileMap[y * mMapWidth + x];
                tile.base = layer[0];
                tile.lights = layer[1];
                if (tile.lights && (tile.lights & 1) == 0)
                    ++mLevelLightCount;
            }
        }
    }
    delete[] tempMap;

    mTileImage = mGraphics->loadImage(tileImageFilename);
    if (mTileImage == NULL)
        success = FALSE;

    if (!success)
        destroyMap();

    return success;
}


void GameState::destroyMap()
{
    delete mTileImage;
    delete[] mTileMap;
    mTileMap = NULL;
    mTileImage = NULL;
}


void GameState::onActivate(UINT32 /*time*/)
{
    mLevels.resetGame();
    mLevels.startLevel();
}


void GameState::startLevel(FLOAT scrollSpeed)
{
    mState = STATE_RUNNING;
    mTime = 0;
    mScrollPos = 0;
    mScrollSpeed = scrollSpeed;
    mBullets = RELOAD_BULLETS;
    mLightsKilled = 0;
    mBulletsShot = 0;
    mUI.deactivateAllComponents();
    mNotifyContinueFlag = FALSE;
    mReloadTime = 0;
}


BOOL GameState::update(UINT32 time)
{
    if (mNotifyContinueFlag)
    {
        mLevels.handleContinue();
        mNotifyContinueFlag = FALSE;
        return FALSE;
    }

    if (mTime != 0)
        mScrollPos += (time - mTime) * mScrollSpeed;
    else
        mStartTime = time;
    mTime = time;

    if (mScrollPos >= mMapWidth * TILE_WIDTH - WIDTH)
    {
        if (mState != STATE_FINISHED)
        {
            mState = STATE_FINISHED;
            mLevels.notifyLevelFinished();
            mUI.activateAllComponents();
        }
        mScrollPos = mMapWidth * TILE_WIDTH - WIDTH;
    }

    if (mReloadTime > 0 && mTime - mReloadTime >= GUN_RELOAD_TIME)
    {
        mBullets = RELOAD_BULLETS;
        mReloadTime = 0;
    }
    mParticleSystem.update();

    return TRUE;
}


void GameState::reload()
{
    mBullets = 0;
    //mBullets = RELOAD_BULLETS;
    mReloadTime = mTime;
    playSoundEffect(SOUND_GUN_RELOAD);
}


void GameState::render(SDL_Surface *screen)
{
    assert(mMapWidth > 0 && mMapHeight > 0);
    assert(mTileMap != NULL);
    assert(mTileImage != NULL);

    // clear with ambient color
    glClearColor(0.15f, 0.15f, 0.15f, 1);
    glClear(GL_COLOR_BUFFER_BIT);

    mGraphics->beginRender();

    renderLevel();
    mParticleSystem.render();

    if (mState == STATE_RUNNING)
        renderBulletSystem();

    mLevels.renderScore();
    if (mTime - mStartTime < LEVEL_START_TEXT_TIME)
        mLevels.renderStartText();
    if (mState == STATE_FINISHED)
        mLevels.renderFinishedText();

    mGraphics->endRender();

    mUI.render();
}


void GameState::renderLevel()
{
    glPushAttrib(GL_COLOR_BUFFER_BIT);

    INT32 x, y;
    FLOAT imageWidth = (FLOAT)((INT32)mTileImage->getWidth());
    FLOAT imageHeight = (FLOAT)((INT32)mTileImage->getHeight());
    FLOAT tileUVWidth = (FLOAT)TILE_WIDTH / imageWidth;
    FLOAT tileUVHeight = (FLOAT)TILE_HEIGHT / imageHeight;
    INT32 tilesPerRow = mTileImage->getWidth() / TILE_WIDTH;
    INT32 tileXStart = (INT32)floor(mScrollPos / TILE_WIDTH);
    INT32 tileXEnd = tileXStart + WIDTH / TILE_WIDTH;
    if (tileXEnd >= mMapWidth)
        tileXEnd = mMapWidth - 1;
    INT32 tileXStartLights = tileXStart - TILE_LIGHT_RENDERING_EXTRA_BORDER;
    INT32 tileXEndLights = tileXEnd + TILE_LIGHT_RENDERING_EXTRA_BORDER;
    if (tileXStartLights < 0)
        tileXStartLights = 0;
    if (tileXEndLights >= mMapWidth)
        tileXEndLights = mMapWidth - 1;
    
    // draw additive light areas
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    for (y = 0; y < mMapHeight; ++y)
    {
        for (x = tileXStartLights; x <= tileXEndLights; ++x)
        {
            Tile &tile = mTileMap[y * mMapWidth + x];
            INT32 tileIndex = tile.lights;
            if (tileIndex && (tileIndex & 1) == 0)
            {
                INT32 lightImageIndex = (tileIndex >> 1) & (MAX_LIGHT_IMAGES - 1);
                const Light &light = mLights[lightImageIndex];
                UIImageOpenGL *lightImage = (UIImageOpenGL *)light.mImage;
                const FLOAT *uv = lightImage->getUV();
                FLOAT xOffset = light.mXOffset, yOffset = light.mYOffset;
                FLOAT cx = -mScrollPos + x * TILE_WIDTH + TILE_WIDTH / 2;
                FLOAT cy = y * TILE_HEIGHT + TILE_HEIGHT / 2;
                FLOAT width = light.mWidth, height = light.mHeight;
                cx += xOffset;
                cy += yOffset;
                glBindTexture(GL_TEXTURE_2D,
                              ((UIImageOpenGL *)lightImage)->getGLId());
                glBegin(GL_QUADS);
                glTexCoord2f(uv[0], uv[1]);
                glVertex2f(cx, cy);
                glTexCoord2f(uv[2], uv[1]);
                glVertex2f(cx + width, cy);
                glTexCoord2f(uv[2], uv[3]);
                glVertex2f(cx + width, cy + height);
                glTexCoord2f(uv[0], uv[3]);
                glVertex2f(cx, cy + height);

                glEnd();
            }
        }
    }

    // tile texture
    glBindTexture(GL_TEXTURE_2D, ((UIImageOpenGL *)mTileImage)->getGLId());

    // multiply in the tile image
    glBlendFunc(GL_ZERO, GL_SRC_COLOR);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBegin(GL_QUADS);
    for (y = 0; y < mMapHeight; ++y)
    {
        for (x = tileXStart; x <= tileXEnd; ++x)
        {
            Tile &tile = mTileMap[y * mMapWidth + x];
            INT32 tileIndex = tile.base;
            INT32 tileRow = tileIndex / tilesPerRow;
            INT32 tileColumn = tileIndex % tilesPerRow;
            glTexCoord2f(tileColumn * tileUVWidth, tileRow * tileUVHeight);
            glVertex2f(-mScrollPos + x * TILE_WIDTH, y * TILE_HEIGHT);
            glTexCoord2f((tileColumn + 1) * tileUVWidth, tileRow * tileUVHeight);
            glVertex2f(-mScrollPos + (x + 1) * TILE_WIDTH, y * TILE_HEIGHT);
            glTexCoord2f((tileColumn + 1) * tileUVWidth, (tileRow + 1) * tileUVHeight);
            glVertex2f(-mScrollPos + (x + 1) * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
            glTexCoord2f(tileColumn * tileUVWidth, (tileRow + 1) * tileUVHeight);
            glVertex2f(-mScrollPos + x * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
        }
    }
    glEnd();

    // draw additive light images
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glBegin(GL_QUADS);
    for (y = 0; y < mMapHeight; ++y)
    {
        for (x = tileXStart; x <= tileXEnd; ++x)
        {
            Tile &tile = mTileMap[y * mMapWidth + x];
            INT32 tileIndex = tile.lights;
            if (tileIndex)
            {
                INT32 tileRow = tileIndex / tilesPerRow;
                INT32 tileColumn = tileIndex % tilesPerRow;
                glTexCoord2f(tileColumn * tileUVWidth, tileRow * tileUVHeight);
                glVertex2f(-mScrollPos + x * TILE_WIDTH, y * TILE_HEIGHT);
                glTexCoord2f((tileColumn + 1) * tileUVWidth, tileRow * tileUVHeight);
                glVertex2f(-mScrollPos + (x + 1) * TILE_WIDTH, y * TILE_HEIGHT);
                glTexCoord2f((tileColumn + 1) * tileUVWidth, (tileRow + 1) * tileUVHeight);
                glVertex2f(-mScrollPos + (x + 1) * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
                glTexCoord2f(tileColumn * tileUVWidth, (tileRow + 1) * tileUVHeight);
                glVertex2f(-mScrollPos + x * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
            }
        }
    }
    glEnd();

    glPopAttrib();
}


void GameState::renderBulletSystem()
{
    if (mBullets == 0 && mReloadTime == 0)
    {
        UIFont *font = GameApp::getInstance()->getMediumFont();

        glDisable(GL_TEXTURE_2D);
        FLOAT alpha = (FLOAT)(sin(mTime * 0.01) * 0.5 + 0.5);
        glColor4f(1, 0.9f, 0.25f, alpha);
        glBegin(GL_QUADS);
        glVertex2f(0, HEIGHT - RELOAD_AREA_HEIGHT);
        glVertex2f(WIDTH, HEIGHT - RELOAD_AREA_HEIGHT);
        glVertex2f(WIDTH, HEIGHT);
        glVertex2f(0, HEIGHT);
        glEnd();
        glEnable(GL_TEXTURE_2D);

        font->drawString(WIDTH / 2, HEIGHT - 1,
                         UIFont::ALIGN_BOTTOM | UIFont::ALIGN_HORIZ_CENTER,
                         "RELOAD");
    }
    else
    {
        INT32 a;
        for (a = 0; a < mBullets; ++a)
            mBulletImage->draw((FLOAT)(WIDTH - mBulletImage->getWidth() * (a + 1)),
                               (FLOAT)(HEIGHT - mBulletImage->getHeight()) - 4);
    }
}


void GameState::onKeyDown(const SDLKey &key)
{
    if (key == SDLK_SPACE && mReloadTime == 0)
        reload();
}


void GameState::onMouseDown(const SDL_MouseButtonEvent &event)
{
    if (mUI.handleUIPointerDown(event.x, event.y))
        return;

    if (mState == STATE_FINISHED)
        return;

    if (event.button == SDL_BUTTON_RIGHT && mReloadTime == 0)
    {
        reload();
        return;
    }
    if (mBullets == 0)
    {
        playSoundEffect(SOUND_GUN_CLICK);
        return;
    }

    INT32 mapPosX = (INT32)(event.x + mScrollPos);
    INT32 mapPosY = event.y;
    INT32 tileX = mapPosX / TILE_WIDTH;
    INT32 tileY = mapPosY / TILE_HEIGHT;
    assert(tileX >= 0 && tileX < mMapWidth);
    assert(tileY >= 0 && tileY < mMapHeight);
    INT32 inTileX = mapPosX % TILE_WIDTH;
    INT32 inTileY = mapPosY % TILE_HEIGHT;
    Tile &tile = mTileMap[tileY * mMapWidth + tileX];
    if (tile.lights && (tile.lights & 1) == 0 &&
        inTileX >= TILE_WIDTH / 4 && inTileX < TILE_WIDTH * 3 / 4 &&
        inTileY >= TILE_HEIGHT / 4 && inTileY < TILE_HEIGHT * 3 / 4)
    {
        INT32 lightIndex = tile.lights;
        Light &light = mLights[(lightIndex >> 1) & (MAX_LIGHT_IMAGES - 1)];
        mLevels.addScore((INT32)(light.mScore * mScrollSpeed * 20));
        ++tile.lights;
        ++mLightsKilled;
        playSoundEffect(SOUND_CLING);
        mParticleSystem.spawn(event.x, event.y,
                              -mScrollSpeed * 1000 / mFPS, 0);
    }
    playSoundEffect(SOUND_GUN);
    --mBullets;
    ++mBulletsShot;
}


void GameState::onMouseUp(const SDL_MouseButtonEvent &event)
{
    if (mUI.handleUIPointerUp(event.x, event.y))
        return;
}


void GameState::onMouseMotion(const SDL_MouseMotionEvent &event)
{
    if (mUI.handleUIPointerMove(event.x, event.y))
        return;

    if (mBullets == 0 && event.y >= HEIGHT - RELOAD_AREA_HEIGHT &&
        mReloadTime == 0)
        reload();
}


void GameState::onAction(UIComponent *component)
{
    if (component == mContinueButton)
        mNotifyContinueFlag = TRUE;
}
