/* Copyright 2005 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: ParticleSystem.cpp,v 1.1 2005/04/17 16:39:18 tonic Exp $
 * $Revision: 1.1 $
 */

#include <stdlib.h>
#include "main.h"
#include "ParticleSystem.h"


#define PARTICLE_GRAVITY        (0.1f)
#define SPAWN_PARTICLES         40
#define RANDOM_SPREAD_RADIUS_X  (1.0f)
#define RANDOM_SPREAD_RADIUS_Y  (2.0f)
#define INITIAL_AGE_RANDOMNESS  90
#define PARTICLE_AGE_SCALE      100
#define INITIAL_YI              -2


static FLOAT clamp(FLOAT value, FLOAT min, FLOAT max)
{
    return value < min ? min : (value > max ? max : value);
}


static FLOAT step(FLOAT value, FLOAT stepPosition)
{
    return value < stepPosition ? 0 : 1;
}


static FLOAT boxStep(FLOAT value, FLOAT slopeStart, FLOAT slopeEnd)
{
    FLOAT diff = slopeEnd - slopeStart;
    if (diff == 0)
        return step(value, slopeStart);
    return clamp((value - slopeStart) / diff, 0, 1);
}


ParticleSystem::ParticleSystem() :
    mParticleImage(NULL)
{
}


ParticleSystem::~ParticleSystem()
{
    delete mParticleImage;
}


BOOL ParticleSystem::init()
{
    UIGraphics *g = GameApp::getInstance()->getGraphics();
    mParticleImage = g->loadImage(DATA_PREFIX "flare-small.png");
    if (mParticleImage == NULL)
        return FALSE;
    return TRUE;
}


void ParticleSystem::spawn(FLOAT x, FLOAT y, FLOAT xi, FLOAT yi)
{
    INT a;
    FLOAT rScale = 0.2f + 0.3f * (rand() / (FLOAT)RAND_MAX);
    FLOAT gScale = (0.4f + 0.5f * (rand() / (FLOAT)RAND_MAX)) * (1 - rScale);
    FLOAT gAdd = 0.2f + 0.1f * (rand() / (FLOAT)RAND_MAX);
    FLOAT bScale = 0.5f * (rand() / (FLOAT)RAND_MAX);
    for (a = 0; a < SPAWN_PARTICLES; ++a)
    {
        Particle *p = new Particle;
        p->x = x;
        p->y = y;
        FLOAT rx, ry;
        do 
        {
            rx = rand() / (FLOAT)RAND_MAX;
            ry = rand() / (FLOAT)RAND_MAX;
            rx = rx * 2 - 1;
            ry = ry * 2 - 1;
        }
        while (rx * rx + ry * ry > 1);
        rx *= RANDOM_SPREAD_RADIUS_X;
        ry *= RANDOM_SPREAD_RADIUS_Y;
        p->xi = xi + rx;
        p->yi = yi + ry + INITIAL_YI;
        p->age = rand() % INITIAL_AGE_RANDOMNESS;
        FLOAT r = rand() / (FLOAT)RAND_MAX;
        p->color[0] = (1 - rScale) + rScale * r;
        p->color[1] = gAdd + gScale * r;
        p->color[2] = bScale * r;
        p->color[3] = 1;
        mParticles.push_back(p);
    }
}


void ParticleSystem::update()
{
    ParticleList::iterator it;
    INT32 maxY = HEIGHT + mParticleImage->getHeight() / 2;

    it = mParticles.begin();
    while (it != mParticles.end())
    {
        Particle *p = *it;
        p->x += p->xi;
        p->y += p->yi;
        p->yi += PARTICLE_GRAVITY;
        ++p->age;
        if (p->y > maxY)
        {
            delete p;
            it = mParticles.erase(it);
        }
        else
            ++it;
    }
}


void ParticleSystem::render()
{
    glPushAttrib(GL_COLOR_BUFFER_BIT);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    UIImageOpenGL *image = (UIImageOpenGL *)mParticleImage;
    glBindTexture(GL_TEXTURE_2D, image->getGLId());
    FLOAT *uv = image->getUV();
    FLOAT imgSize = (FLOAT)image->getWidth();

    glBegin(GL_QUADS);
    ParticleList::iterator it;
    it = mParticles.begin();
    for (; it != mParticles.end(); ++it)
    {
        Particle *p = *it;
        /*
        mParticleImage->draw(UIImage::ADD,
                             UIImage::TRANSFORM | UIImage::CENTER,
                             p->x, p->y);
        */
        glColor4fv(p->color);
        FLOAT x = p->x, y = p->y;
        FLOAT s = (imgSize - boxStep(p->age, 0, PARTICLE_AGE_SCALE) * imgSize * 0.85f) / 2;
        glTexCoord2f(uv[0], uv[1]); glVertex2f(x - s, y - s);
        glTexCoord2f(uv[2], uv[1]); glVertex2f(x + s, y - s);
        glTexCoord2f(uv[2], uv[3]); glVertex2f(x + s, y + s);
        glTexCoord2f(uv[0], uv[3]); glVertex2f(x - s, y + s);
    }
    glEnd();

    glPopAttrib();
}
