/* Copyright 2005 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: sound.cpp,v 1.5 2005/04/17 22:50:54 tonic Exp $
 * $Revision: 1.5 $
 */

#include "fmod.h"
#include "main.h"
#include "sound.h"


//#define DISABLE_SOUND


#define CHANNELS    32
#define MUSICFILE   DATA_PREFIX "lightwar.oxm"
#define MUSICVOLMUL 0.75f
#define SFXVOLMUL   1.0f


typedef struct
{
    INT type;
    UINT8 volume;
    char *filename;
} SOUNDEFF;


static SOUNDEFF sampleData[] =
{
    { SOUND_GUN, 0x80, DATA_PREFIX "gun.ogg" },
    { SOUND_GUN_CLICK, 0xff, DATA_PREFIX "click.ogg" },
    { SOUND_GUN_RELOAD, 0xff, DATA_PREFIX "reload.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling1.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling2.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling3.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling4.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling5.ogg" },
    { SOUND_CLING, 0xf0, DATA_PREFIX "cling6.ogg" },
};
#define SAMPLES (sizeof(sampleData) / sizeof(sampleData[0]))


static BOOL soundInitialized = FALSE, soundEnabled = TRUE, musicEnabled = FALSE;
static FSOUND_SAMPLE *samples[SAMPLES];
static FMUSIC_MODULE *music;
static int musicChn = -1;
static float musicVolume = 1.0f;


static FSOUND_SAMPLE * loadSample(char *fileName)
{
#ifndef DISABLE_SOUND
    return FSOUND_Sample_Load(FSOUND_FREE, fileName, FSOUND_LOOP_OFF, 0, 0);
#else
    return NULL;
#endif // DISABLE_SOUND
}


static void loadSamples()
{
#ifndef DISABLE_SOUND
    int a;

    for (a = 0; a < SAMPLES; ++a)
    {
        samples[a] = loadSample(sampleData[a].filename);
        if (samples[a] == NULL)
        {
            errorMessage("Sample Load Error", "Error loading sample %s\nError: %s",
                         sampleData[a].filename, SDL_GetError());
            exit(1);
        }
    }
#endif // !DISABLE_SOUND
}


static void loadMusic()
{
#ifndef DISABLE_SOUND
    music = FMUSIC_LoadSong(MUSICFILE);
    if (music == NULL)
    {
        errorMessage("Music Load Error", "Error loading file %s\nError code: %d",
                     MUSICFILE, FSOUND_GetError());
        musicEnabled = 0;
    }
#endif // !DISABLE_SOUND
}


void initSound()
{
#ifndef DISABLE_SOUND
    soundInitialized = 1;
    if (!FSOUND_Init(44100, CHANNELS, 0))
    {
        errorMessage("Sound Disabled", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
        soundInitialized = soundEnabled = 0;
        return;
    }

    loadSamples();
    loadMusic();
#endif // !DISABLE_SOUND
}


void deinitSound()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    setMusic(0, 0);

    UINT a;
    for (a = 0; a < SAMPLES; ++a)
    {
        FSOUND_Sample_Free(samples[a]);
        samples[a] = NULL;
    }
    FMUSIC_FreeSong(music);

    FSOUND_Close();
#endif // !DISABLE_SOUND
}


int playSoundEffect(int type, int forceChannel)
{
#ifndef DISABLE_SOUND
    int chn, snd, cnt = 0;

    if (!soundInitialized || !soundEnabled)
        return -1;

    do {
        snd = rand() % SAMPLES;
        ++cnt;
    } while (sampleData[snd].type != type && cnt < 10000);
    if (cnt >= 10000)
        return -1; // don't play anything if we can't find a sample
        //errorMessage("Sound Error", "Unable to find sample");

    chn = FSOUND_PlaySound(FSOUND_FREE, samples[snd]);

    if (chn != -1)
        FSOUND_SetVolume(chn, sampleData[snd].volume);

    return chn;
#else
    return 0;
#endif // DISABLE_SOUND
}


void stopSoundEffect(int channel)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized || channel < 0)
        return;
    FSOUND_StopSound(channel);
#endif // DISABLE_SOUND
}


void stopAllSoundEffects()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    int a;
    for (a = 0; a < CHANNELS; ++a)
    {
        if (musicEnabled && musicChn == a)
            continue;   // don't stop music if music is enabled
        stopSoundEffect(a);
    }
#endif // DISABLE_SOUND
}


void setSound(BOOL enabled)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    soundEnabled = enabled ? 1 : 0;
#endif // !DISABLE_SOUND
}


void setMusic(BOOL enabled, INT order)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    if (enabled && !musicEnabled)
    {
        // turn music on
        musicEnabled = FMUSIC_PlaySong(music) ? 1 : 0;
        setMusicVolume(musicVolume);
    }
    else if (musicEnabled && !enabled)
    {
        // turn music off
        FMUSIC_StopAllSongs();
        musicEnabled = 0;
    }
    if (musicEnabled)
        FMUSIC_SetOrder(music, order);

#endif // !DISABLE_SOUND
}


void setMusicVolume(float volume)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    musicVolume = volume;

    if (musicEnabled)
        FMUSIC_SetMasterVolume(music, (int)(volume * MUSICVOLMUL * 256));
#endif // !DISABLE_SOUND
}


BOOL getSoundEnabled()
{
#ifndef DISABLE_SOUND
    return soundEnabled;
#else
    return FALSE;
#endif
}


BOOL getMusicEnabled()
{
#ifndef DISABLE_SOUND
    return musicEnabled;
#else
    return FALSE;
#endif
}
